import os, itertools, math, statistics
import numpy as np
from collections import defaultdict
from collections import OrderedDict
from sortedcollections import ValueSortedDict
from natsort import natsorted

doAll = True

# Function to extract statistics from a line
def extract_stat(line):
    prefix, value = line.split(":")
    clean_prefix = " ".join(prefix.split())  # Remove extra spaces
    err = 0
    value = value.replace("+/-", "±")
    if "±" in value:
        err = float(value.strip().split("±")[1].rstrip("%").strip())
    return clean_prefix, (float(value.strip().split()[0].rstrip("%")), err)


def compute_stat(line):
    n_str, result_str = line.split("\t", 1)
    result_str = result_str.split("\t")[0]
    n = float(n_str)
    result = float(result_str)
    correct=round(n*(result/100.0))
    p=correct/n
    err=((p*(1-p)/(n-1))**0.5)*100.0
    return (float(result), err)

def avg(a, b):
    return (a + b) / 2.0

# Directories containing the result and evaluation files
resultsDirectory = "results"
evaluationDirectory = "evaluation"

models = set()
perplexity_stats = defaultdict(lambda: defaultdict(dict))
kl_divergence_stats = defaultdict(lambda: defaultdict(dict))
token_probability_stats = defaultdict(lambda: defaultdict(dict))
arc_easy_stats = defaultdict(dict)
arc_challenge_stats = defaultdict(dict)
mmlu_stats = defaultdict(dict)
winogrande_stats = defaultdict(dict)
eval_stats = defaultdict(dict)


# Iterate over each file in the directories
for entry in itertools.chain(os.scandir(resultsDirectory), os.scandir(evaluationDirectory)):
    if not entry.name.endswith(".txt"):
        continue
    if not doAll:
        if "i1" in entry.name:
            if "IQ1_S" in entry.name:
                continue
            if "IQ1_M" in entry.name:
                continue
        else:
            if "IQ3_XS" in entry.name:
                continue
            if "IQ3_S" in entry.name:
                continue
            if "IQ3_M" in entry.name:
                continue
    print(f"Processing file: {entry.path}")
    
    # Extract the base name and quant from the file name
    if entry.path.startswith("evaluation"):
        splits = entry.name.rsplit('.', 3)
    else:
        splits = entry.name.rsplit('.', 2)
    model = splits[0]
    quant = splits[1]
    models.add(model)
    
    with open(entry.path, encoding="utf-8") as file:
        #print(entry.path)
        if entry.name.endswith(".arc-easy.txt"):
            for line in file:
                if line.startswith("570\t"):
                    arc_easy_stats[model][quant] = compute_stat(line)
        elif entry.name.endswith(".arc-challenge.txt"):
            for line in file:
                if line.startswith("299\t"):
                    arc_challenge_stats[model][quant] = compute_stat(line)
        elif entry.name.endswith(".mmlu.txt"):
            for line in file:
                if line.startswith("1548\t"):
                    mmlu_stats[model][quant] = compute_stat(line)
        elif entry.name.endswith(".winogrande.txt"):
            for line in file:
                if line.startswith("1267\t"):
                    winogrande_stats[model][quant] = compute_stat(line)
        else:
            for line in file:
                if ":" in line and not "multiple_choice_score" in line:
                    key, value = extract_stat(line)
                    if "PPL" in key:
                        perplexity_stats[key][model][quant] = value
                    elif "KLD" in key:
                        kl_divergence_stats[key][model][quant] = value
                    elif "Δp" in key or "Same top p" in key:
                        token_probability_stats[key][model][quant] = value


model = None
for modelItem in natsorted(models):
    if "-Instruct" not in modelItem:
        continue
    if "0.5B" in modelItem:
        continue
    instructModel = modelItem
    baseModel = modelItem.replace("-Instruct", "")
    if "-72B" in modelItem:
        baseModel = instructModel
    kl_divergence_ranking = ValueSortedDict()
    token_probability_correct_ranking = ValueSortedDict()
    token_probability_same_token_ranking = ValueSortedDict()
    perplexity_ranking  = ValueSortedDict()
    eval_ranking  = ValueSortedDict()
    for quant in kl_divergence_stats["Mean KLD"][instructModel].keys():
        
        #if quant in ["Q4_0_8_8", "Q4_0_4_8"]:
        #    continue
        
        #print(modelItem, quant)
        
        normArcEasyBase = (arc_easy_stats[baseModel][quant][0]/arc_easy_stats[baseModel]["SOURCE"][0]) * 570
        normArcChallengeBase = (arc_challenge_stats[baseModel][quant][0]/arc_challenge_stats[baseModel]["SOURCE"][0]) * 299
        normMMLUBase = (mmlu_stats[baseModel][quant][0]/mmlu_stats[baseModel]["SOURCE"][0]) * 1548
        normWinograndeBase = (winogrande_stats[baseModel][quant][0]/winogrande_stats[baseModel]["SOURCE"][0]) * 1267
        wAvgEvalBase = (normArcEasyBase + normArcChallengeBase + normMMLUBase + normWinograndeBase)/3684.0
        wAvgEvalPercentBase = wAvgEvalBase * 100
        
        normArcEasyInstruct = (arc_easy_stats[instructModel][quant][0]/arc_easy_stats[instructModel]["SOURCE"][0]) * 570
        normArcChallengeInstruct = (arc_challenge_stats[instructModel][quant][0]/arc_challenge_stats[instructModel]["SOURCE"][0]) * 299
        normMMLUInstruct = (mmlu_stats[instructModel][quant][0]/mmlu_stats[instructModel]["SOURCE"][0]) * 1548
        normWinograndeInstruct = (winogrande_stats[instructModel][quant][0]/winogrande_stats[instructModel]["SOURCE"][0]) * 1267
        wAvgEvalInstruct = (normArcEasyInstruct + normArcChallengeInstruct + normMMLUInstruct + normWinograndeInstruct)/3684.0
        wAvgEvalPercentInstruct = wAvgEvalInstruct * 100
        
        kl_divergence_base = 100 - kl_divergence_stats["Mean KLD"][baseModel][quant][0] * 100
        token_probability_correct_base = token_probability_stats["Mean Δp"][baseModel][quant][0] + 100
        token_probability_same_token_base = token_probability_stats["Same top p"][baseModel][quant][0]
        perplexityBase = 100 + (100 - perplexity_stats["Mean PPL(Q)/PPL(base)"][baseModel][quant][0] * 100)
        
        kl_divergence_instruct = 100 - kl_divergence_stats["Mean KLD"][instructModel][quant][0] * 100
        token_probability_correct_instruct = token_probability_stats["Mean Δp"][instructModel][quant][0] + 100
        token_probability_same_token_instruct = token_probability_stats["Same top p"][instructModel][quant][0]
        perplexityInstruct = 100 + (100 - perplexity_stats["Mean PPL(Q)/PPL(base)"][instructModel][quant][0] * 100)
        
        kl_divergence_ranking[quant] = -avg(kl_divergence_base, kl_divergence_instruct)
        token_probability_correct_ranking[quant] = -avg(token_probability_correct_base, token_probability_correct_instruct)
        token_probability_same_token_ranking[quant] = -avg(token_probability_same_token_base, token_probability_same_token_instruct)
        perplexity_ranking[quant] = -avg(perplexityBase, perplexityInstruct)
        eval_ranking[quant] = -avg(wAvgEvalPercentBase, wAvgEvalPercentInstruct)
        
    kl_divergence_ranking_pos = OrderedDict()
    for i, quant in enumerate(kl_divergence_ranking):
        kl_divergence_ranking_pos[quant] = i+1
    
    token_probability_correct_ranking_pos = OrderedDict()
    for i, quant in enumerate(token_probability_correct_ranking):
        token_probability_correct_ranking_pos[quant] = i+1
    
    token_probability_same_token_ranking_pos = OrderedDict()
    for i, quant in enumerate(token_probability_same_token_ranking):
        token_probability_same_token_ranking_pos[quant] = i+1
    
    perplexity_ranking_pos = OrderedDict()
    for i, quant in enumerate(perplexity_ranking):
        perplexity_ranking_pos[quant] = i+1
    
    eval_ranking_pos = OrderedDict()
    for i, quant in enumerate(eval_ranking):
        eval_ranking_pos[quant] = i+1
    
    #print(kl_divergence_ranking_pos)
    #print(token_probability_correct_ranking_pos)
    #print(token_probability_same_token_ranking_pos)
    #print(perplexity_ranking_pos)
    #print(eval_ranking_pos)
    
    quality = ValueSortedDict()
    for quant in kl_divergence_stats["Mean KLD"][instructModel].keys():
        
        #if quant in ["Q4_0_8_8", "Q4_0_4_8"]:
        #    continue
        
        quality[quant] = kl_divergence_ranking_pos[quant] * 0.25 + token_probability_correct_ranking_pos[quant] * 0.40 + token_probability_same_token_ranking_pos[quant] * 0.20 + perplexity_ranking_pos[quant] * 0.15
    #print(quality)
    
    quality_pos = OrderedDict()
    print()
    print(f"## {baseModel} & {instructModel}")
    print("| Quant | Rank | KL-divergence | Correct token | Same token | Perplexity | Eval |")
    print("| -- | -- | -- | -- | -- | -- | -- |")
    for i, quant in enumerate(quality):
        rank = i + 1
        quality_pos[quant] = rank
        print(f'| {quant} | {rank} | {-kl_divergence_ranking[quant]:.2f} | {-token_probability_correct_ranking[quant]:.2f} | {-token_probability_same_token_ranking[quant]:.2f} | {-perplexity_ranking[quant]:0.2f} | {-eval_ranking[quant]:0.2f} |')
    input()
